@@ -12,8 +12,14 @@
 #include "output.h" // dprintf
 #include "std/vbe.h" // struct vbe_info
 #include "string.h" // memset_far
-#include "vgabios.h" // handle_104f
+#include "vgabios.h" // get_current_mode
 #include "vgahw.h" // vgahw_set_mode
+#include "vgautil.h" // handle_104f
+
+#define VBE_OEM_STRING "SeaBIOS VBE(C) 2011"
+#define VBE_VENDOR_STRING "SeaBIOS Developers"
+#define VBE_PRODUCT_STRING "SeaBIOS VBE Adapter"
+#define VBE_REVISION_STRING "Rev. 1"
 
 u32 VBE_total_memory VAR16 = 256 * 1024;
 u32 VBE_capabilities VAR16;
@@ -217,7 +223,7 @@ vbe_104f02(struct bregs *regs)
 static void
 vbe_104f03(struct bregs *regs)
 {
-    regs->bx = GET_BDA(vbe_mode);
+    regs->bx = GET_BDA_EXT(vbe_mode);
     dprintf(1, "VBE current mode=%x\n", regs->bx);
     regs->ax = 0x004f;
 }
@@ -247,7 +253,7 @@ vbe_104f05(struct bregs *regs)
 {
     if (regs->bh > 1 || regs->bl > 1)
         goto fail;
-    if (GET_BDA(vbe_mode) & MF_LINEARFB) {
+    if (GET_BDA_EXT(vbe_mode) & MF_LINEARFB) {
         regs->ah = VBE_RETURN_STATUS_INVALID;
         return;
     }
@@ -382,10 +388,10 @@ vbe_104f10(struct bregs *regs)
         regs->bx = 0x0f30;
         break;
     case 0x01:
-        SET_BDA(vbe_flag, regs->bh);
+        MASK_BDA_EXT(flags, BF_PM_MASK, regs->bh & BF_PM_MASK);
         break;
     case 0x02:
-        regs->bh = GET_BDA(vbe_flag);
+        regs->bh = GET_BDA_EXT(flags) & BF_PM_MASK;
         break;
     default:
         regs->ax = 0x014f;
